/******************************************/
/* GCommand - Run command using execvp    */
/*            Bypassing /bin/sh           */
/* Author -   Alan Swanson                */
/*            30/06/02                    */
/******************************************/

/******************************************/
/* This program is free software; you can redistribute it and/or
/* modify it under the terms of the GNU General Public License
/* as published by the Free Software Foundation; either version 2
/* of the License, or (at your option) any later version.
/* 
/* This program is distributed in the hope that it will be useful,
/* but WITHOUT ANY WARRANTY; without even the implied warranty of
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/* GNU General Public License for more details.
/* 
/* You should have received a copy of the GNU General Public License
/* along with this program; if not, write to the Free Software
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
/******************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "gcommand.h"

/******************************************/
/* Run command                            */
/******************************************/

extern pid_t
g_command(gchar * command)
{
	pid_t pid;
	gchar **argv;
	gint argc;

	/* Divide command line into a argv/argc pair */
	if (g_command_parse(command, &argv, &argc))
		return 0;

	if ((pid = fork()) == 0) {
		execvp(argv[0], argv);

		exit(EXIT_FAILURE);
	}

	free(argv);

	return pid;
}

/******************************************/
/* Parse command                          */
/******************************************/

extern gint
g_command_parse(gchar * command, gchar *** argv, gint * argc)
{
	GSList *list = NULL;
	gchar *ptr;
	gchar *option, *cur;
	guint i, quote = 0;

	ptr = command;

	if ((option = malloc(strlen(command + 1))) == NULL)
		return 1;

	/* Overflow if command not terminated */
	while (*ptr) {
		/* Skip leading space */
		while (*ptr == ' ' || *ptr == '\t')
			ptr++;

		cur = option;

		while (1) {
			/* Escaped   - Force character                 */
			/* Space/Tab - End of option if outside quotes */
			/* Quotes    - Valid characters within quotes  */
			if (*ptr == '\\')
				ptr++;
			else if ((*ptr == ' ' || *ptr == '\t') && !quote) {
				*cur = 0;
				break;
			} else if (*ptr == '"' && quote != 1) {
				quote = !quote * 2;
				ptr++;
				continue;
			} else if (*ptr == '\'' && quote != 2) {
				quote = !quote;
				ptr++;
				continue;
			}

			/* Break at any null */
			if (*ptr == 0) {
				*cur = 0;
				break;
			}

			/* Valid character */
			*cur = *ptr;
			cur++;
			*cur = 0;
			ptr++;
		}

		/* Copy individual command options to linked list */
		if (*option)
			list = g_slist_append(list, g_strdup(option));

		*option = (gchar) NULL;
	}

	free(option);

	/* Initialise argv/argc pair */
	if ((*argc = g_slist_length(list)) == 0)
		return 1;
	if ((*argv = malloc(sizeof (gchar *) * (*argc + 1))) == NULL)
		return 1;

	/* Copy data from linked list into array          */
	/* Last element of argv must be NULL              */
	/* NB g_slist_nth_data returns NULL past end list */
	for (i = 0; i <= *argc; i++) {
		(*argv)[i] = g_slist_nth_data(list, i);
	}

	/* Free linked list memory */
	g_slist_free(list);

	return 0;
}
